/*
 * dvdplayer.c: DVD player for DVBs PCI card
 *
 * Copyright (C) 2000 Matjaz Thaler
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * Or, point your browser to http://www.gnu.org/copyleft/gpl.html
 * 
 * The author can be reached at matjaz.thaler@rd.iskraemeco.si 
 *
 * the project's page is at http://www2.arnes.si/~mthale1/index.html
 */


#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>

#include "css-descramble.h"
#include "css-init.h"
#include "subtitles.h"

void usage();
void skipAC3(FILE *dev,char *buffer, int *stat);
void synchronize(int file, unsigned char *buffer);

unsigned char title_key[5];
unsigned char sector[2048];

int video=1;            // 1 playback video
int css=0;              // 1 css decode, 
int ac3=-2;             // -2 no sound, -1 if only 1 soundtrack, 0-7 play soundtrack x
int ac3stat=0;          // 0 normal, 1 skip frame
int inputFile=0;        // stdin
int skip=0;             // skip MB of input file
int subtitles=-1;       // 0-31 subtitles x
int ac3sync=-1;         // -1 normal, 1-xx skip frames 
char fileName[256];
int videoDev;
FILE *ac3pipe;
int spuPipe;

int main(int argc, char **argv){
  int c, i;
  int ac3count=0;
  
  while(1){
    c = getopt(argc, argv, "va:cf:g:s:j:h::");
    if (c == -1)
      break;
    switch (c){
    case 'v':
      video=0;
      fprintf (stderr, "do not playback video \n");
      break;
    case 'a':
      if(optarg[0]=='a'){
	ac3=-1;
	fprintf (stderr, "ac3 sound\n");
      }
      else{
	sscanf(optarg, "%d", &ac3);
	ac3 &= 7;
	fprintf (stderr, "ac3 soundtrack %d\n", ac3);
      }
      break;
    case 'c':
      css=1;
      fprintf (stderr, "css decode\n");
      break;
    case 'f':
      strcpy(fileName, optarg);
      inputFile=-1;
      break;
    case 'g':
      sscanf(optarg, "%d", &skip);
      fprintf (stderr, "skip %d MB\n", skip);
      break;
    case 's':
      sscanf(optarg, "%d", &subtitles);
      subtitles &= 31;
      fprintf (stderr, "subtitles %d\n", subtitles);
      break;
    case 'j':
      sscanf(optarg, "%d", &ac3sync);
      fprintf (stderr, "ac3skip %d\n", ac3sync);
      ac3sync=4096/ac3sync;
      break;
    case 'h':
    default:
      usage();
      return(1);
    }
  }
                                                            // init 
  if((video==1) || (subtitles >= 0)){
    videoDev=open("/dev/video", O_RDWR);
    if(videoDev < 0){
      fprintf (stderr, "can't open /dev/video\n");
      return(1);
    }
  }
  if(subtitles >= 0){
    if(!fork()){
      spuPipe=open("/tmp/spu", O_RDONLY);
      if(spuPipe < 0){
	fprintf(stderr, "can't open /dev/spu\n");
	exit(1);
      }
      subtitles_decode(spuPipe, videoDev);
      close(spuPipe);
      exit(0);
    }
    spuPipe=open("/tmp/spu", O_WRONLY);
    if(spuPipe < 0){
      fprintf(stderr, "can't open /dev/spu\n");
      exit(1);
    }
  }
  if(inputFile!=0){
    inputFile=open(fileName, O_RDONLY);
    if(inputFile<0){
      fprintf (stderr, "can't open %s\n", fileName);
      return(1);
    }
    fprintf (stderr, "input file: %s\n", fileName);
  }
  lseek (inputFile, skip*1024*1024, SEEK_CUR);
  if(css==1)
    css_init(title_key);
  if(ac3 != -2){
    ac3pipe=popen("ac3dec", "w");
    if(ac3pipe==NULL){
      fprintf (stderr, "can't run ac3dec\n");
      return(1);
    }
  }
  while(1){                                                    //playback
    if(read(inputFile, sector, 2048) == 2048){
      if (!(sector[0]==0x00 && sector[1]==0x00 && sector[2]==0x01 && sector[3]==0xba)) {
	synchronize(inputFile, sector);
      }
      if ((sector[14+3]&0xF0)==0xE0 || (sector[14+3]&0xF0)==0xC0){ // video, mpeg audio
	if(video){
	  if ((sector[20] & 0x30) && css==1) {                 // css decode
	    css_descramble(sector, title_key);
	    sector[20] &= 0x8f;
	  }
	  write(videoDev, sector , 2048);
	}
      }
      else if (((ac3>=0) && ((sector[14+sector[14+8]+9])==(0x80|ac3))) || 
	       ((ac3==-1) && ((sector[14+sector[14+8]+9]&0xF8)==0x80))) {  // Dolby AC3 audio
	if ((sector[20] & 0x30) && css==1) {                 // css decode
	  css_descramble(sector, title_key);
	  sector[20] &= 0x8f;
	}
	ac3count++;
	if((ac3count%ac3sync || ac3sync<1) && ac3stat==0)
	  fwrite((sector+0x20), (2048-0x20), 1, ac3pipe);
	else                                                 // skip ac3frame
	  skipAC3(ac3pipe, sector, &ac3stat);
      }
      else if ((subtitles>=0) && (sector[14+sector[14+8]+9])==(0x20|subtitles)) { //spu data
	if ((sector[20] & 0x30) && css==1) {                 // css decode
	  css_descramble(sector, title_key);
	  sector[20] &= 0x8f;
	}
	i = 24 + sector[22];
	write(spuPipe, sector+i , 2048-i);
      }
    }
    else{                                                   // end of file
      int len=strlen(fileName);
      if(inputFile!=0){
	close(inputFile);
	if(fileName[len-5]>'0' && fileName[len-5]<'9' && len>=12){
	  fileName[len-5]++;                                // find next file
	  inputFile=open(fileName, O_RDONLY);              
	  if(inputFile<0){
	    fprintf (stderr, "can't open %s\n", fileName);
	    break;
	  }
	  else
	    fprintf (stderr, "playing: %s\n", fileName);  
	}
	else break;
      }
      else break;
    }
  }
  if(subtitles >= 0){
    close(spuPipe);
    sleep(2);
  }
  if(ac3 != -2)
    fclose(ac3pipe);
  if((video==1) || (subtitles >= 0))
    close(videoDev);
  return 0;
} 


void usage(){
  fprintf (stderr, "\n");
  fprintf (stderr, "Usage:\tdvdplayer [OPTION]...\n");
  fprintf (stderr, "OPTIONS:\n");
  fprintf (stderr, " -f filename\topen file (default: stdin)\n");
  fprintf (stderr, " -c\t\tcss decode (playback from DVD)\n");
  fprintf (stderr, " -a channel\tplay ac3 sound with ac3dec\n");
  fprintf (stderr, " -s subtitle  \tdisplay subtitles\n");
  fprintf (stderr, " -v\t\tdo not play video\n");
  fprintf (stderr, " -g n\t\tstart playback at n M bytes of input file\n");
  fprintf (stderr, " -j n\t\tskip n/4096 ac3 frames -");
  fprintf (stderr, " use this option if sound has delay\n");
  fprintf (stderr, " -h\t\tshow this usage information\n");
  fprintf (stderr, "\n");
  fprintf (stderr, "EXAMPLE:\n");
  fprintf (stderr, " -playback from DVD:\n");
  fprintf (stderr, "   dvdplayer -a 0 -c -s 0 -f /dvd/video_ts/vts_01_1.vob\n");
  fprintf (stderr, " -playback from hard disk:\n");
  fprintf (stderr, "   dvdplayer -a 0 -s 0 -f foo.vob\n");
  fprintf (stderr, "\n");
}

void synchronize(int file, unsigned char *buffer){
  int sync=0;
  
  fprintf(stderr, "sync\n");
  while(sync==0){
    while(buffer[0] != 0)
      read(file, &buffer[0], 1);
    read(file, &buffer[1], 1);
    if(buffer[1] != 0)
      continue;
    buffer[2]=0;
    while(buffer[2] == 0)
      read(file, &buffer[2], 1);
    if(buffer[2] != 1)
      continue;
    read(file, &buffer[3], 1);
    if(buffer[3] != 0xba)
      continue;
    read(file, &buffer[4], 2044);
    sync=1;
  }
}

void skipAC3(FILE *dev, char *buffer, int *stat){
  int i=0x20,j,last=0;

  if(*stat==0)                 // 0 normal, 1 skip frame
    *stat=1;
  while(!(buffer[i]==0x0b && buffer[i+1]==0x77)){
    if(i++>486){
      fwrite((buffer+0x20), 2048-0x20, 1, dev);
      return;
    }
  }
  for(j=i+256; j<2048; j+=256){
    if(buffer[j]==0x0b && buffer[j+1]==0x77){
      last=1;
    }
  }
  j-=256;
  if(last==0)
    fwrite((buffer+0x20), 2048-0x20, 1, dev);
  else{
    fwrite((buffer+0x20), i-0x20, 1, dev);
    fwrite((buffer+j), 2048-j, 1, dev);
    *stat=0;
    //printf("skip frame: %3.3d %3.3d\n", i, j);
  }
  return;
}
